#!/usr/local/bin/ruby
#-*-Ruby-*-
# $Id: MHC.rb,v 1.2 1999/12/02 19:18:16 oharakun Exp $

################################################################
###
### Classes for handling schedule articles.
###
################################################################
class ScheduleDB
  SCH_DB_HASH = {}

  MON = %w(jan feb mar apr may jun jul aug sep oct nov dec)
  WEK = %w(sun mon tue wed thu fri sat)
  ORD = %w(1st 2nd 3rd 4th 5th last)

  def initialize(yyyy = nil, mm = nil)
    @yyyy, @mm = yyyy, mm
  end

  def to_slot_a(yyyy, mm)
    ScheduleDB .to_slot_a(yyyy, mm)
  end

  def ScheduleDB.to_slot_a(yyyy, mm)
    home      = ENV['HOME']
    rcfile    = home + '/.schedule'
    intersect = home + '/Mail/schedule/intersect'
    yyyymm    = sprintf("#{home}/Mail/schedule/%04d/%02d", yyyy, mm)
    return [rcfile, intersect, yyyymm]
  end

  def scan(yyyy, mm)
    ScheduleDB .scan(yyyy, mm)
  end

  def ScheduleDB.scan(yyyy, mm)
    to_slot_a(yyyy, mm) .each{|path|
      if (modified?(path))
	SCH_DB_HASH[path] = {}
	ScheduleDB .update(path)
      end
    }
  end

  def update(path)
    ScheduleDB .update(path)
  end

  def ScheduleDB .update(path)
    if (File .file?(path))
      ## if path is a ordinary file, it must be a rc file.
      file = File .open(path, "r")

      while(header = file .gets("\n\n"))
	header .gsub!(/\n\s+/, ' ')
	if (header =~ /^[^#]/)
	  hdr = {}
	  header .split("\n") .each{|line|
	    if (line =~ /^X-SC-([^:]+:)\s*(.*)/i)
	      key, val = $1 .downcase, $2 .strip
	      hdr[key] = val if (val != '')
	    end
	  }
	  i = ScheduleItem .new(path, hdr, path + ':' + ($. .to_s))
	end
      end
    elsif (File .directory?(path))
      ## path must be a directory, yyyymm folder.

      Dir.glob(path + '/*') .each {|file_path|
	hdr = {}
	begin
	  file = File .open(file_path, "r")
	  header = file .gets("\n\n")
	rescue
	  printf "Invalid File #{file_path} ignored.\n"
	end
	header .gsub!(/\n\s+/, ' ')
	header .split("\n") .each{|line|
	  if (line =~ /^X-SC-([^:]+:)\s*(.*)/i)
	    key, val = $1 .downcase, $2 .strip
	    hdr[key] = val if (val != '')
	  end
	}
	ScheduleItem .new(path, hdr, file_path)
      }
    end
  end

  def modified?(slot)
    ScheduleDB .modified?(slot)
  end
  
  def ScheduleDB.modified?(slot)
    return false if (!File .exists?(slot))
    return true  if (SCH_DB_HASH[slot] .nil?)
    return true  if (SCH_DB_HASH[slot]['mtime'] .nil?)

    if (SCH_DB_HASH[slot]['mtime'] < File .mtime(slot))
      return true
    else
      return false
    end
  end

  def modified_any?(yyyy, mm)
    ScheduleDB .modified_any?(yyyy, mm)    
  end

  def ScheduleDB.modified_any?(yyyy, mm)
    to_slot_a(yyyy, mm) .each{|slot|
      return true if (ScheduleDB .modified?(slot))
    }
    return false
  end
  
  def search(yyyy = @yyyy, mm = @mm, dd = @dd)
    ScheduleDB .search(yyyy, mm, dd)
  end

  def ScheduleDB.search(yyyy, mm, dd)
    mon = MON[mm - 1]
    wek = WEK[day_of_week(yyyy, mm, dd)]
    ord = ORD[(dd - 1) / 7]
    ret = []
    all = 'all'; last = 'last'
    day = sprintf("%02d", dd)
    yyyymmdd = sprintf("%04d%02d%02d", yyyy, mm, dd)

    search_key = [yyyymmdd, mon+ord+wek, mon+all+wek, all+ord+wek,
      all+all+wek, mon+day, all+day]

    if (dd + 7 > days_of_month(yyyy, mm))
      search_key << mon+last+wek << all+last+wek
    end

    ScheduleDB .to_slot_a(yyyy, mm) .each{|slot|
      search_key .each{|key|
	if (!(SCH_DB_HASH[slot] .nil?) and SCH_DB_HASH[slot][key] .is_a?(Array))
	  SCH_DB_HASH[slot][key] .each{|item|
	    if !(item .exc .include?('!' + yyyymmdd) or
		 (item .dur_s != '' and
		  item .dur_e != '' and
		  (yyyymmdd < item .dur_s or yyyymmdd > item .dur_e)))
	      ret << item
	    end
	  }
	end
      }
    }
    return ret .sort{|a,b| a .tim_s <=> b .tim_s} .uniq
  end

  def ScheduleDB.regist(slot, key, obj)
    if (SCH_DB_HASH[slot] .nil?)
      SCH_DB_HASH[slot] = {}
    end

    if (SCH_DB_HASH[slot][key] .nil?)
      SCH_DB_HASH[slot][key] = [obj]
    else
      SCH_DB_HASH[slot][key] << obj
    end

    SCH_DB_HASH[slot]['mtime'] = Time .now .localtime
  end

  def regist(slot, key, obj)
    ScheduleDB .regist(slot, key, obj)
  end

  ################################################################
  ## schedule item class
  ##
  class ScheduleItem
  ################################################################
  ## Class constants
  ################################################################
  MON = %w(jan feb mar apr may jun jul aug sep oct nov dec)
  WEK = %w(sun mon tue wed thu fri sat)
  ORD = %w(1st 2nd 3rd 4th 5th last)

  MON_REGEX = MON .join('|')
  WEK_REGEX = WEK .join('|')
  ORD_REGEX = ORD .join('|')

  ################################################################
  ## access methods to instance variables
  ################################################################
  attr :mon         # array of string MON or all
  attr :ord         # array of string ORD or all
  attr :wek         # array of string WEK or all
  attr :num         # array of string dd
  attr :day         # array of string yyyymmdd
  attr :exc         # array of string yyyymmdd
  attr :cat         # array of string
  attr :tim_s       # string hh:mm or ''
  attr :tim_e       # string hh:mm or ''
  attr :dur_s       # string yyyymmdd
  attr :dur_e       # string yyyymmdd
  attr :sub         # string
  attr :path        # string

  ################################################################
  ## initialize
  ################################################################
  def initialize(slot, hdr, path)
    @mon, @ord, @wek, @num, @day, @exc, @cat = [], [], [], [], [], [], []
    @sub, @tim_s, @tim_e, @dur_s, @dur_e = '', '', '', '', ''
    @path = path
    @hdr = hdr

    ## regist obj to DB.
    @hdr .each_pair{|key,val|
      case key
      when 'day:'
	val .split .each{|d|
	  case d
	  when /^!\d+$/
	    @exc << d
	  when /^\d+$/
	    @day << d
	    regist(slot, d, self)
	  end
	}
      when 'date:'
	## backward compatibility
        if (val =~ /(\d+)\s+([A-Z][a-z][a-z])\s+(\d+)\s+(\d\d:\d\d)/)
	  dd, mm, yy, hhmm = $1 .to_i, $2, $3 .to_i  + 1900 , $4
	  mm = ("JanFebMarAprMayJunJulAugSepOctNovDec" .index(mm)) / 3 + 1
	  yyyymmdd = sprintf("%04d%02d%02d", yy, mm, dd)
	  @tim_s = (hhmm != '00:00')? hhmm : ''
	  @day << yyyymmdd
	  regist(slot, yyyymmdd, self)
	end
      when 'subject:'
	@sub = val

      when 'time:'
	@tim_s, @tim_e = val .split('-')
	@tim_e = @tim_e .to_s

      when 'duration:'
	@dur_s, @dur_e = val .split('-')
	@dur_e = @dur_e .to_s

      when 'category:'
	val .split .each{|c|
	  @cat << c .downcase
	}
      when 'cond:'
	val .split .each{|d|
	  case d
	  when /^(#{MON_REGEX})$/oi
	    @mon << d .downcase
	  when /^(#{ORD_REGEX})$/oi
	    @ord << d .downcase
	  when /^(#{WEK_REGEX})$/oi
	    @wek << d .downcase
	  when /^\d+$/
	    @num << sprintf("%02d", d .to_i)
	  end
	}
	@mon << 'all' if (@mon .empty?)
	@ord << 'all' if (@ord .empty?)
	@mon .each{|mon|
	  @ord .each{|ord|
	    @wek .each{|wek|
	      regist (slot, mon+ord+wek, self)
	    }
	  }
	  @num .each{|dd|
	    regist(slot, mon+dd, self)
	  }
	}
      end ## case ##
    }
  end

  private
  def regist(slot, key, obj)
    ScheduleDB .regist(slot, key, obj)
  end
  end
end


################################################################
###
### common date functions
###
################################################################

################################################################
## year:
##    ex. 1994
## return: 
##    true  ... leap year
##    false ... not leap year
##
def leap?(year)
  year % 4 == 0 and year % 100 != 0 or year % 400 == 0
end

################################################################
## year:
##    ex. 1994
## return: 
##    1 ... leap year
##    0 ... not leap year
##
def leap(year)
  return (leap?(year)) ? 1 : 0
end

################################################################
## Nth day of the year.
##
## mon, day, year:
##    ex. 7, 1, 1994
## return:
##    182 ... 7, 1, 1994 is the 182nd day of the year.
##
def day_of_year(year, mon, day)
  return (day + (mon - 1) * 31 -
	  [0, 0, 3, 3, 4, 4, 5, 5, 5, 6, 6, 7][mon - 1] + 
	  ((leap?(year) and mon > 2)? 1:0))
end

################################################################
## days of the month.
##
## mon, year:
##    ex. 2, 1992
## return:
##    29 (1992 is leap year.)
##
def days_of_month(year, mon)
  [31, 28 + leap(year), 31, 30, 31, 30,
   31, 31, 30, 31, 30, 31][mon - 1];
end

################################################################
## day of week as a number.
##
## mon, day, year:
##    ex. 7, 1, 1994
## return:
##    5 (Fri) ... 7, 1, 1994 is Friday (0:Sun, 1:Mon, ... , 6:Sat)
##
def day_of_week(year, mon, day)
  y = year - 1
  return (day_of_year(year, mon, day) + y*365 + y/4 - y/100 + y/400) % 7
end
